local L = LibStub("AceLocale-3.0"):GetLocale("Critline")
local playerGUID, hasSpecialAura, inVehicle

--[[ global addon variables ]]
CritlineDebug = false -- for internal testing only, leave it set to false!

SUBHEADER_TEXT_COLOR  = "|cffCEA208"
CRITLINE_DMG_ICON = "Interface\\Icons\\Ability_SteelMelee"
CRITLINE_HEAL_ICON = "Interface\\Icons\\Spell_Holy_FlashHeal"
CRITLINE_PET_ICON = "Interface\\Icons\\Ability_Hunter_Pet_Bear"

CL_CRIT_TEXT = TEXT_MODE_A_STRING_RESULT_CRITICAL:match("(%a+)") or TEXT_MODE_A_STRING_RESULT_CRITICAL

Critline_OutOfControl = false
Critline_MindControlled = false


local addon = CreateFrame("Frame", "Critline")
addon:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("PLAYER_LOGIN")
addon:RegisterEvent("PLAYER_CONTROL_LOST")
addon:RegisterEvent("PLAYER_CONTROL_GAINED")
addon:RegisterEvent("PLAYER_DEAD")
addon:SetScript("OnEvent", function(self, event, ...)
	if self[event] then
		return self[event](self, event, ...)
	end
end)

addon.summary = {dmg = "", heal = "", pet = ""}


addon.splash = CreateFrame("MessageFrame", "CritlineSplashFrame", UIParent)
addon.splash.locked = true
addon.splash:SetFrameStrata("LOW")
addon.splash:SetMovable(true)
addon.splash:RegisterForDrag("LeftButton")
addon.splash:SetWidth(512)
addon.splash:SetHeight(96)
addon.splash:SetPoint("CENTER")
addon.splash:SetFontObject("NumberFontNormalHuge")
addon.splash:SetScript("OnUpdate", function(self)
	if not self.locked then
		self:AddMessage(L["Critline splash frame unlocked"], CritlineSettings.spellColor)
		self:AddMessage(L["Drag to move"], CritlineSettings.amountColor)
		self:AddMessage(L["Right-click to lock"], CritlineSettings.amountColor)
	end
end)
addon.splash:SetScript("OnMouseUp", function(self, button)
	if button == "RightButton" then
		self.locked = true
		self:EnableMouse(false)
		self:Clear()
	end
end)
addon.splash:SetScript("OnDragStart", function(self)
	if not self.locked then
		self:StartMoving()
	end
end)
addon.splash:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
end)

addon.splash.OldAddMessage = addon.splash.AddMessage
function addon.splash:AddMessage(msg, color, ...)
	self:OldAddMessage(msg, color.r, color.g, color.b, ...)
end


function addon:SettingsReset()
	CritlineSettings = {}
	self:Initialize()
end


function addon:ResetAll()
	CritlineSettings = {}
	CritlineDB = {}
	self.summary = {dmg = "", heal = "", pet = ""}
	self:Initialize()
end


function addon:GetHighest(tree, hitType)
	-- hittype is 'NORM' or 'CRIT'
	-- tree is 'DMG', 'PET' or 'HEAL'
	
	local hidmg = 0
	local spell = "n/a"
	
	-- if not (CritlineDB and CritlineDB[tree]) then 
		-- return hidmg, spell
	-- end

	for k, v in pairs(CritlineDB[tree]) do
		if v[hitType] and v[hitType].amount > hidmg then
			if self.filters:SpellPassesFilter(tree, k) then
				hidmg = v[hitType].amount
				spell = k
			end
		end
	end
	return hidmg, spell
end


function addon:RebuildAllTooltips()
	self.summary = {dmg = "", heal = "", pet = ""}	-- have to rebuild cached data when a spell is filtered to update tooltips
	self:GetSummaryRichText("dmg") 
	self:GetSummaryRichText("heal")
	self:GetSummaryRichText("pet")
	self:DoUpdate()
end


function addon:GetSummaryRichText(tree)
	-- This is for code optimization. We only build the rich summary when a new record is hit.
	-- Then a new record is recorded, the engine will blank out the Critline_Summary[tree]
	-- This causes the tree to be rebuilt when this function is called.
	-- tree is 'DMG', 'PET' or 'HEAL'
	
	if #self.summary[tree] == 0 then
		self.summary[tree] = self:BuildSummaryRichText(tree)
	end
	return self.summary[tree]
end


local recordSorters = {
	-- alpha: sort by name
	alpha = function(a, b)
		return a.spellName < b.spellName
	end,
	-- crit: sort by crit > normal > name
	crit = function(a, b)
		if (a.crit and a.crit.amount or 0) == (b.crit and b.crit.amount or 0) then
			if (a.normal and a.normal.amount or 0) == (b.normal and b.normal.amount or 0) then
				return a.spellName < b.spellName
			else
				return (a.normal and a.normal.amount or 0) > (b.normal and b.normal.amount or 0)
			end
		else
			return (a.crit and a.crit.amount or 0) > (b.crit and b.crit.amount or 0)
		end
	end,
	-- normal: sort by normal > crit > name
	normal = function(a, b)
		if (a.normal and a.normal.amount or 0) == (b.normal and b.normal.amount or 0) then
			if (a.crit and a.crit.amount or 0) == (b.crit and b.crit.amount or 0) then
				return a.spellName < b.spellName
			else
				return (a.crit and a.crit.amount or 0) > (b.crit and b.crit.amount or 0)
			end
		else
			return (a.normal and a.normal.amount or 0) > (b.normal and b.normal.amount or 0)
		end
	end,
}


local trees = {dmg = "damage", heal = "healing", pet = "pet"}
local sortedSpells = {}

function addon:BuildSummaryRichText(tree)
	local rtfAttack = ""
	local normalRecord = self:GetHighest(tree, "normal")
	local critRecord = self:GetHighest(tree, "crit")
	local commonFormat = "$sn\n"
	local normalFormat = "   "..L["Normal"]..": $na\t $nt ($nl)\n"
	local critFormat = "   "..CL_CRIT_TEXT..": $ca\t $ct ($cl)\n"

	-- attempt to dismantle user provided tooltip format
	local tooltipFormat
	if CritlineSettings.tooltipFormat ~= "" then
		tooltipFormat = CritlineSettings.tooltipFormat.."\n"
		tooltipFormat = tooltipFormat:gsub("\\t", "\t")
		tooltipFormat = tooltipFormat:gsub("\\n", "\n")
		commonFormat = ""
		normalFormat = ""
		critFormat = ""
		for line in tooltipFormat:gmatch("[^\n]+") do
			local normalMatch = line:match("$n[atl]")
			local critMatch = line:match("$c[atl]")
			if normalMatch and critMatch then
				commonFormat = commonFormat..line.."\n"
			elseif normalMatch then
				normalFormat = line.."\n"
			elseif critMatch then
				critFormat = line.."\n"
			else
				commonFormat = line.."\n"
			end
		end
	end

	-- $sn - Spell name

	-- $na - Amount for normal record
	-- $nt - Target name for normal record
	-- $nl - Target level for normal record

	-- $ca - Amount for crit record
	-- $ct - Target name for crit record
	-- $cl - Target level for crit record
	
	wipe(sortedSpells)
	
	for k, v in pairs(CritlineDB[tree]) do
		tinsert(sortedSpells, {
			spellName = k,
			normal = v.normal,
			crit = v.crit,
		})
	end
	
	sort(sortedSpells, recordSorters[CritlineSettings.summarySort])
	
	local formattedEntry
	local normal, normalAmount, normalTarget, normalTargetLevel
	local crit, critAmount, critTarget, critTargetLevel

	for _, data in pairs(sortedSpells) do
		if self.filters:SpellPassesFilter(tree, data.spellName) then
			normal = normalFormat
			crit = critFormat
			
			-- get stored normal data
			if data.normal then
				if data.normal.amount == normalRecord then
					normalAmount = GREEN_FONT_COLOR_CODE..data.normal.amount..FONT_COLOR_CODE_CLOSE
				else
					normalAmount = HIGHLIGHT_FONT_COLOR_CODE..data.normal.amount..FONT_COLOR_CODE_CLOSE
				end
				normalTarget = data.normal.target
				normalTargetLevel = data.normal.targetLevel
			else
				normal = ""
				normalAmount = HIGHLIGHT_FONT_COLOR_CODE.."0"..FONT_COLOR_CODE_CLOSE
				normalTarget = ""
				normalTargetLevel = ""
			end

			-- get stored crit data
			if data.crit then
				if data.crit.amount == critRecord then
					critAmount = GREEN_FONT_COLOR_CODE..data.crit.amount..FONT_COLOR_CODE_CLOSE
				else
					critAmount = HIGHLIGHT_FONT_COLOR_CODE..data.crit.amount..FONT_COLOR_CODE_CLOSE
				end
				critTarget = data.crit.target or ""
				critTargetLevel = data.crit.targetLevel or ""
			else
				crit = ""
				critAmount = HIGHLIGHT_FONT_COLOR_CODE.."0"..FONT_COLOR_CODE_CLOSE
				critTarget = ""
				critTargetLevel = ""
			end

			if CritlineSettings.detailedTooltip then
				formattedEntry = commonFormat..normal..crit
			else
				formattedEntry = "$sn\t[$na/$ca]\n"
			end

			-- substitute tags with database values
			formattedEntry = formattedEntry:gsub("$sn", data.spellName)

			formattedEntry = formattedEntry:gsub("$na", normalAmount)
			formattedEntry = formattedEntry:gsub("$nt", HIGHLIGHT_FONT_COLOR_CODE..normalTarget..FONT_COLOR_CODE_CLOSE)
			formattedEntry = formattedEntry:gsub("$nl", normalTargetLevel)

			formattedEntry = formattedEntry:gsub("$ca", critAmount)
			formattedEntry = formattedEntry:gsub("$ct", HIGHLIGHT_FONT_COLOR_CODE..critTarget..FONT_COLOR_CODE_CLOSE)
			formattedEntry = formattedEntry:gsub("$cl", critTargetLevel)

			rtfAttack = rtfAttack..formattedEntry
		end
	end

	if rtfAttack == "" then
		return "|cffffffffCritline "..trees[tree].."\n"..L["No records"]
	end
	
	return strsub("|cffffffffCritline "..trees[tree].."\n"..rtfAttack, 1, -2)
end


function addon:ADDON_LOADED(event, addon)
	if addon == "Critline" then
		self:Initialize()
		self.splash:EnableMouse(0)
		self.splash:Clear()
		self.loaded = true
	end
end


function addon:PLAYER_LOGIN()
	playerGUID = UnitGUID("player")
	self:Debug("playerGUID: "..playerGUID)
	
	inVehicle = UnitInVehicle("player")
	
	if self.filters:HasSpecialAura() then
		hasSpecialAura = true
		self:Debug("Special aura found. Hits won't be recorded.")
	end
end


function addon:COMBAT_LOG_EVENT_UNFILTERED(event, ...)
	if not self.loaded then
		return
	end
	
	local timestamp, eventType, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags = ...
	local isPet
	-- we seem to get events with standard arguments equal to nil, so they need to be ignored
	if not (timestamp and eventType) then
		self:Debug("nil errors on start")
		return
	end

	-- if we don't have a destName (who we hit or healed) and we don't have a sourceName (us or our pets) then we leave
	if not (destName or sourceName) then
		self:Debug("nil source/dest")
		return
	end
	
	if destGUID == playerGUID then
		local spellID = select(9, ...)
		if (eventType == "SPELL_AURA_APPLIED" or eventType == "SPELL_AURA_REMOVED" or eventType == "SPELL_AURA_BROKEN" or eventType == "SPELL_AURA_BROkEN_SPELL") then
			for i, v in ipairs(Critline_SpecialAuras) do
				if v == spellID then
					if self.filters:HasSpecialAura() then
						hasSpecialAura = true
						self:Debug("Special aura detected. Suppressing record tracking.")
					else
						hasSpecialAura = false
						self:Debug("No special aura detected. Resuming record tracking.")
					end
					return
				end
			end
		end
	end
	
	if hasSpecialAura then
		self:Debug("Special aura found. Return.")
		return
	end
	
	local myPetGUID = UnitGUID("pet")
	
	-- if sourceGUID is not us or our pet, we leave
	if sourceGUID ~= playerGUID then
		if myPetGUID then
			-- Critline_Debug("myPetGUID: "..myPetGUID)
			if sourceGUID ~= myPetGUID then
				-- Critline_Debug("sourceGUID("..sourceGUID..") not us("..playerGUID..") or our pet("..myPetGUID..")")
				return
			elseif inVehicle then
				self:Debug("In a vehicle, ignore.")
				return
			end
		else
			-- Critline_Debug("sourceGUID ("..sourceGUID..") ~= playerGUID ("..playerGUID.."). Return.")
			return
		end
	end
	
	isPet = sourceGUID == myPetGUID

	if (Critline_MindControlled and CritlineSettings.suppressMC) then
		-- we are suppressing mind control events, time to leave
		self:Debug("Mind Control suppressed")
		return
	end
	
	if (eventType == "SPELL_CAST_START" or eventType == "SPELL_CAST_FAILED" or eventType == "SPELL_CAST_SUCCESS" or eventType == "SWING_DAMAGE") then
		if Critline_OutOfControl then
			-- if we are out of control and casting spells and/or attacking, we are mind controlled
			Critline_MindControlled = true
		end
	end
	if eventType == "SWING_DAMAGE" then
		-- can be non-Physical damage (e.g. Melee that hits as Shadow damage)
		local amount, _, school, _, _, _, critical = select(9, ...)
		if (amount and amount > 0) and not addon.filters:IsMobInFilter(destName, destGUID) then -- feyde..check for mob in filter
			self:HitHandle(destGUID, destName, "Melee", false, critical, amount, school, isPet)
		end
	elseif (eventType == "RANGE_DAMAGE" or eventType == "SPELL_DAMAGE" or eventType == "SPELL_PERIODIC_DAMAGE" or eventType == "DAMAGE_SPLIT" or eventType == "DAMAGE_SHIELD") then
		local _, spellName, _, amount, _, school, _, _, _, critical = select(9, ...)
		if (amount and amount > 0) and not addon.filters:IsMobInFilter(destName, destGUID) then -- feyde..check for mob in filter
			self:HitHandle(destGUID, destName, spellName, false, critical, amount, school, isPet)
		end
	-- healing events
	elseif (eventType == "SPELL_HEAL" or eventType == "SPELL_PERIODIC_HEAL") then
		local _, spellName, _, amount, _, critical = select(9, ...)
		if (amount and amount > 0) and not self.filters:IsMobInFilter(destName, destGUID) then -- feyde..check for mob in filter
			self:HitHandle(destGUID, destName, spellName, true, critical, amount, 0, isPet)
		end
	end
end


function addon:UNIT_ENTERED_VEHICLE()
	inVehicle = true
end


function addon:UNIT_EXITED_VEHICLE()
	inVehicle = false
end


function addon:PLAYER_CONTROL_LOST()
	if not self.loaded then
		return
	end
	-- this covers fear, taxi, mind control, and others
	Critline_OutOfControl = true
end


function addon:PLAYER_CONTROL_GAINED()
	if not self.loaded then
		return
	end
	Critline_OutOfControl = false
	Critline_MindControlled = false
end


function addon:PLAYER_DEAD()
	hasSpecialAura = self.filters:HasSpecialAura()
end


function addon:Initialize()
	self:Debug("Initializing...")

	-- settings
	CritlineSettings = CritlineSettings or {}
	
	local defaultSettings = {
		-- SETTING = DefaultValue
		firstLoad = true,
		showMinimap = true,
		minimapPos = 0,
		playSound = true,
		screenshot = false,
		dmgRecord = true,
		healRecord = true,
		petRecord = true,
		dmgDisplay = true,
		healDisplay = true,
		petDisplay = true,
		dmgInvert = false,
		healInvert = false,
		petInvert = false,
		recordPvE = true,
		recordPvP = true,
		suppressMC = false,
		levelAdjust = -1,
		detailedTooltip = false,
		summarySort = "crit",
		tooltipFormat = "",
		dontFilterMagic = false,
		noSelfDamage = false,
		showSplash = true,
		sctSplash = false,
		mainScale = 1,
		splashScale = 1,
		splashTimer = 2,
		spellColor = {
			r = 1,
			g = 1,
			b = 0,
		},
		amountColor = {
			r = 1,
			g = 1,
			b = 1,
		},
	}
	
	for setting, value in pairs(defaultSettings) do
		if CritlineSettings[setting] == nil then
			CritlineSettings[setting] = value
		end
	end
	
	-- records database
	if not CritlineDB then
		CritlineDB = {
			dmg = {},
			heal = {},
			pet = {},
		}
		if CL_DATABASE then
			self:Debug("Converting old database...")
			for tree, treeData in pairs(CL_DATABASE) do
				tree = tree:lower()
				for spell, spellData in pairs(treeData) do
					CritlineDB[tree][spell] = {
						normal = spellData.NORM,
						crit = spellData.CRIT,
					}
					for _, data in pairs(spellData) do
						if type(data) ~= "string" then -- prevent error from corrupt database
							data.amount = data.Amount
							data.target = data.Target
							data.targetLevel = data.Level
							data.Amount = nil
							data.Target = nil
							data.Level = nil
						end
					end
				end
			end
		end
		self:Debug("Database created.")
	end
	
	-- repair database since r14, and remove deprecated entries
	for _, tree in pairs(CritlineDB) do
		for _, spell in pairs(tree) do
			spell.spell = nil
			spell.Filter = nil
			if spell.normal and spell.normal.target then
				spell.normal.Count = nil
				spell.normal.Date = nil
			else
				spell.normal = nil
			end
			if spell.crit and spell.crit.target then
				spell.crit.Count = nil
				spell.crit.Date = nil
			else
				spell.crit = nil
			end
		end
	end

	-- filters
	if not CritlineSpellFilter then
		CritlineSpellFilter = {
			dmg = {},
			heal = {},
			pet = {},
		}
		if CL_FILTERS then
			for tree, data in pairs(CL_FILTERS) do
				tree = tree:lower()
				for _, spell in ipairs(data) do
					CritlineSpellFilter[tree][spell] = true
				end
			end
		end
		self:Debug("Spell filter created.")
	end

	if not CritlineMobFilter then
		CritlineMobFilter = {}
		self:Debug("Mob filter created.")
	end
	
	self.splash:SetTimeVisible(CritlineSettings.splashTimer)
	
	self:RebuildAllTooltips()
	self:Debug("Initialization completed.")
end


addon.tooltip = CreateFrame("GameTooltip", "CritlineTooltip", nil, "GameTooltipTemplate")

function addon:GetLevelFromGUID(destGUID)
	self.tooltip:SetOwner(UIParent, "ANCHOR_NONE")
	self.tooltip:SetHyperlink("unit:"..destGUID)
	
	local level = UnitLevel("player") -- by default, all mobs are fair game
	local isPlayer -- by default we are fighting a mob
	
	for i = 1, self.tooltip:NumLines() do
		local mytext = _G["CritlineTooltipTextLeft"..i]
		local text = mytext:GetText()
		if text then
			if text:match(LEVEL) then -- our destGUID has the word Level in it.
				level = text:match("(%d+)")  -- find the level
				if level then  -- if we found the level, break from the for loop
					level = tonumber(level)
				else
					-- well, the word Level is in this tooltip, but we could not find the level
					-- either the destGUID is at least 10 levels higher than us, or we just couldn't find it.
					level = UnitLevel("player")  -- like I said, by default, all mobs are fair game
				end
			end
			if strmatch(text, PLAYER) then -- our destGUID has the word Player in it.
				-- we are fighting another Player, this is used for PvP flag
				isPlayer = true
			end
		end
	end	
	return level, isPlayer
end


function addon:HitHandle(destGUID, destName, spell, isHeal, critical, amount, school, isPet)
	local targetLevel, isPlayer = self:GetLevelFromGUID(destGUID)

	if not (isPlayer or CritlineSettings.recordPvE) or (isPlayer and not CritlineSettings.recordPvP) then
		self:Debug("Target == player/non player and not recording PvP/PvE; return.")
		return
	end
	
	if not isHeal and (CritlineSettings.noSelfDamage and destGUID == playerGUID) then
		self:Debug("Self inflicted damage and ignoring such; return.")
		return
	end

	local leveldiff = 0
	if targetLevel < UnitLevel("player") then
		leveldiff = (UnitLevel("player") - targetLevel)
	end
	
	-- self:Debug("Level difference: "..leveldiff)
	-- self:Debug("Level adjustment: "..CritlineSettings.LVLADJ)
	-- self:Debug("Spell school: "..school)
	
	-- ignore level adjustment if magic damage and the setting is enabled
	if (CritlineSettings.levelAdjust >= 0) and (CritlineSettings.levelAdjust < leveldiff) and (school == 1 or not CritlineSettings.dontFilterMagic) then
		self:Debug("Level adjustment enabled, target level too low and damage school is filtered; return.")
		return
	end

	if not CritlineDB then
		self:Debug("CritlineDB should not be nil at this point!")
		self:Initialize()
	end

	local tree
	if isPet then
		if not isHeal then
			tree = "pet"
		end
	elseif isHeal then
		tree = "heal"
	else
		tree = "dmg"
	end

	-- exit if not recording type dmg.
	if not (tree and CritlineSettings[tree.."Record"]) then
		return
	end
	
	local hitType = critical and "crit" or "normal"

	-- create spell database entries as required
	if not CritlineDB[tree][spell] then
		-- self:Debug("Creating CritlineDB."..tree.."."..spell.."...")
		CritlineDB[tree][spell] = {}
	end

	if not CritlineDB[tree][spell][hitType] then
		self:Debug("Creating CritlineDB."..tree.."."..spell.."."..hitType.."...")
		CritlineDB[tree][spell][hitType] = {}
	end

	if (not CritlineDB[tree][spell][hitType].amount or CritlineDB[tree][spell][hitType].amount < amount) then
		CritlineDB[tree][spell][hitType] = {
			amount = amount,
			target = destName,
			targetLevel = targetLevel,
		}
		addon.announce:Update()	-- update announce..
		addon.reset:Update()	-- ..reset..
		self.filters.spellFilter.scrollFrame:Update()	-- ..and filter list
		self.summary[tree] = ""
		if self.filters:SpellPassesFilter(tree, spell) then
			self:DisplayNewRecord(spell, amount, critical)
		end
	end
end


function addon:DisplayNewRecord(spell, amount, critical)
	if CritlineSettings.showSplash then
		if CritlineSettings.sctSplash then
			if MikSBT then
				if critical then
					MikSBT.DisplayMessage(CL_CRIT_TEXT.."!", nil, true, 255, 0, 0)
				end
				MikSBT.DisplayMessage(format(L["New %s record!"], spell), nil, true, (CritlineSettings.spellColor.r * 255), (CritlineSettings.spellColor.g * 255), (CritlineSettings.spellColor.b * 255))
				MikSBT.DisplayMessage(amount, nil, true, (CritlineSettings.amountColor.r * 255), (CritlineSettings.amountColor.g * 255), (CritlineSettings.amountColor.b * 255))
			elseif SCT then
				if critical then
					SCT:DisplayMessage(CL_CRIT_TEXT.."!", {r = 255, g = 0, b = 0})
				end
				SCT:DisplayMessage(format(L["New %s record!"], spell), CritlineSettings.amountColor)
				SCT:DisplayMessage(amount, CritlineSettings.amountColor)
			elseif SHOW_COMBAT_TEXT == "1" then
				CombatText_AddMessage(amount, CombatText_StandardScroll, CritlineSettings.amountColor.r, CritlineSettings.amountColor.g, CritlineSettings.amountColor.b, "crit")
				CombatText_AddMessage(format(L["New %s record!"], spell), CombatText_StandardScroll, CritlineSettings.spellColor.r, CritlineSettings.spellColor.g, CritlineSettings.spellColor.b, "crit")
				if critical then
					CombatText_AddMessage(CL_CRIT_TEXT.."!", CombatText_StandardScroll, 1, 0, 0, "crit")
				end
			end
		else
			self.splash:Clear()
			if critical then
				self.splash:AddMessage(CL_CRIT_TEXT.."!", {r = 1, g = 0, b = 0})
			end
			self.splash:AddMessage(format(L["New %s record!"], spell), CritlineSettings.spellColor)
			self.splash:AddMessage(amount, CritlineSettings.amountColor)
		end
	end
	if CritlineSettings.playSound then 
		PlaySound("LEVELUP", 1, 1, 0, 1, 3) 
	end
	if CritlineSettings.screenshot then 
		-- To avoid protection issues, you could substitute SetAlpha(0) and SetAlpha(100) for Hide() and Show(), respectively.
		-- if UIParent:IsShown() 
		-- UIParent:Hide() --SetAlpha(0)
		TakeScreenshot() 
		-- UIParent:Show() --SetAlpha(100)
	end
	self:RebuildAllTooltips() --feyde...since Tooltips are now cached, we have to reset them to update the new record data.
end


CLOU_Handles = {}

function addon:DoUpdate()
	for k, v in pairs(CLOU_Handles) do
		v()
	end
end


function addon:OnUpdateRegister(newhandler)
	tinsert(CLOU_Handles, newhandler)
end


--[[ misc help functions ]]
function addon:Debug(msg)
	if CritlineDebug and DEFAULT_CHAT_FRAME then
		DEFAULT_CHAT_FRAME:AddMessage("|cff56a3ffCritlineDebug:|r "..msg)
	end
end


function addon:Message(msg)
	if not msg then
		msg = "Critline_Message passed a nil value."
	end
	if DEFAULT_CHAT_FRAME then
		DEFAULT_CHAT_FRAME:AddMessage("|cffffff00Critline:|r "..msg)
	end
end