local L = LibStub("AceLocale-3.0"):GetLocale("Critline")
local core = Critline
local module

local numSpellFilterButtons = 15
local numMobFilterButtons = 20

local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff

local specialMobs = {
	[12460] = true,	-- Death Talon Wyrmguard
	[12461] = true,	-- Death Talon Overseer
	[14020] = true,	-- Chromaggus
	[15339] = true,	-- Ossirian the Unscarred
	[22841] = true,	-- Shade of Akama
	[15928] = true,	-- Thaddius
}

Critline_SpecialAuras = {
	23620,	-- Burning Adrenaline (Vaelastrasz the Corrupt)
	44406,	-- Energy Infusion (Vexallus)
	44335,	-- Energy Feedback (Vexallus)
	41337,	-- Aura of Anger (Reliquary of Souls)
	41350,	-- Aura of Desire (Reliquary of Souls)
	58549,	-- Tenacity (Lake Wintergrasp)
	60964,	-- Strength of Wrynn (The Battle For The Undercity)
	59641,	-- Warchief's Blessing (The Battle For The Undercity)
	55849,	-- Power Spark (Malygos)
	56648,	-- Potent Fungus (Amanitar)
	56330,	-- Iron's Bane (Storm Peaks quest)
	58026,	-- Blessing of the Crusade (Icecrown quest)
	53642,	-- Might of Mograine (Light's Hope Chapel)
	58361,	-- Might of Mograine (Patchwerk)
}

OldSpellButton_OnModifiedClick = SpellButton_OnModifiedClick


-- template function for spell filter check buttons
local function createSpellButton(self)
	local checkButton = CreateFrame("CheckButton", nil, self, "OptionsBaseCheckButtonTemplate")
	checkButton:SetPushedTextOffset(0, 0)
	checkButton:SetScript("OnClick", function(self)
		CritlineSpellFilterButton_OnClick(self)
		if self:GetChecked() then
			PlaySound("igMainMenuOptionCheckBoxOn")
		else
			PlaySound("igMainMenuOptionCheckBoxOff")
		end
	end)
	
	local checkButtonText = checkButton:CreateFontString(nil, nil, "GameFontHighlight")
	checkButtonText:SetPoint("LEFT", checkButton, "RIGHT", 0, 1)
	checkButton:SetFontString(checkButtonText)
	
	return checkButton
end


-- template function for mob filter buttons
local function createMobButton(self)
	local button = CreateFrame("Button", nil, self)
	button:SetWidth(300)
	button:SetHeight(16)
	button:SetNormalFontObject("GameFontNormal")
	button:SetHighlightFontObject("GameFontHighlight")
	button:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight", "ADD")
	button:SetFontString(button:CreateFontString())
	button:SetPushedTextOffset(0, 0)
	button:SetScript("OnClick", function(self)
		CritlineFiltersButton_OnClick(self)
	end)
	
	return button
end

	
core.filters = CreateFrame("Frame")
module = core.filters
module.name = FILTERS
module.parent = "Critline"
InterfaceOptions_AddCategory(module)

module.spellFilter = CreateFrame("Frame", nil, module)
module.spellFilter:SetWidth(300)
module.spellFilter:SetHeight(numSpellFilterButtons * 22 + 4)
module.spellFilter:SetPoint("TOP", 0, -56)

module.spellFilter.scrollFrame = CreateFrame("ScrollFrame", "CritlineSpellFilterScrollFrame", module.spellFilter, "FauxScrollFrameTemplate")
module.spellFilter.scrollFrame:SetAllPoints()
module.spellFilter.scrollFrame:SetScript("OnShow", function(self)
	FauxScrollFrame_SetOffset(self, 0)
	_G[self:GetName().."ScrollBar"]:SetValue(0)
	self:Update()
end)
module.spellFilter.scrollFrame:SetScript("OnVerticalScroll", function(self, offset) FauxScrollFrame_OnVerticalScroll(self, offset, 22, self.Update) end)

do	-- create spell buttons
	module.spellFilter.buttons = {}
	
	local buttons = module.spellFilter.buttons
	local button
	
	for i = 1, numSpellFilterButtons do
		button = createSpellButton(module.spellFilter)
		if i == 1 then
			button:SetPoint("TOPLEFT")
		else
			button:SetPoint("TOP", buttons[i - 1], "BOTTOM", 0, 4)
		end
		buttons[i] = button
	end
end

module.filterType = CreateFrame("Frame", "CritlineFilterType", module, "UIDropDownMenuTemplate")
module.filterType:SetPoint("BOTTOMLEFT", module.spellFilter, "TOPLEFT", -16, 0)

module.spellFilter.tree = CreateFrame("Frame", "CritlineSpellFilterTree", module.spellFilter, "UIDropDownMenuTemplate")
module.spellFilter.tree:SetPoint("BOTTOMRIGHT", module.spellFilter, "TOPRIGHT", 16, 0)

module.mobFilter = CreateFrame("Frame", nil, module)
module.mobFilter:SetWidth(300)
module.mobFilter:SetHeight(numMobFilterButtons * 16)
module.mobFilter:SetPoint("TOP", 0, -56)
module.mobFilter:Hide()
module.mobFilter:SetScript("OnShow", function(self)
	local scrollFrame = self.scrollFrame
	if scrollFrame.selected then
		local prevHighlight = module.mobFilter.buttons[scrollFrame.selected - FauxScrollFrame_GetOffset(scrollFrame)]
		if prevHighlight then
			prevHighlight:UnlockHighlight()
		end
		self.scrollFrame.selected = nil
		-- self.scrollFrame:Update()
		self.deleteButton:Disable()
	end
end)

do	-- create mob filter buttons
	module.mobFilter.buttons = {}
	for i = 1, numMobFilterButtons do
		module.mobFilter.buttons[i] = createMobButton(module.mobFilter)
		if i == 1 then
			module.mobFilter.buttons[i]:SetPoint("TOP")
		else
			module.mobFilter.buttons[i]:SetPoint("TOP", module.mobFilter.buttons[i - 1], "BOTTOM")
		end
		module.mobFilter.buttons[i]:SetID(i)
	end
end

module.mobFilter.scrollFrame = CreateFrame("ScrollFrame", "CritlineMobFilterScrollFrame", module.mobFilter, "FauxScrollFrameTemplate")
module.mobFilter.scrollFrame:SetAllPoints()
module.mobFilter.scrollFrame:SetScript("OnShow", function(self)
	FauxScrollFrame_SetOffset(self, 0)
	_G[self:GetName().."ScrollBar"]:SetValue(0)
	self:Update()
end)
module.mobFilter.scrollFrame:SetScript("OnVerticalScroll", function(self, offset) FauxScrollFrame_OnVerticalScroll(self, offset, 16, self.Update) end)

module.mobFilter.addTargetButton = CreateFrame("Button", nil, module.mobFilter, "UIPanelButtonTemplate")
module.mobFilter.addTargetButton:SetWidth(122)
module.mobFilter.addTargetButton:SetHeight(24)
module.mobFilter.addTargetButton:SetPoint("TOPLEFT", module.mobFilter, "BOTTOMLEFT")
module.mobFilter.addTargetButton:SetText(L["Add target"])
module.mobFilter.addTargetButton:SetScript("OnClick", function()
	local targetName = UnitName("target")
	if targetName then
		if module:IsMobInFilter(targetName) then
			core:Message(L["%s is already in mob filter."]:format(targetName))
		elseif UnitIsPlayer("target") then
			core:Message(L["Cannot add players to mob filter."])
		else
			tinsert(CritlineMobFilter, targetName)
			module.mobFilter.scrollFrame:Update()
			core:Message(L["%s added to mob filter."]:format(targetName))
		end
	else
		core:Message(L["No target selected."])
	end
end)

module.mobFilter.deleteButton = CreateFrame("Button", nil, module.mobFilter, "UIPanelButtonTemplate")
module.mobFilter.deleteButton:SetWidth(122)
module.mobFilter.deleteButton:SetHeight(24)
module.mobFilter.deleteButton:SetPoint("TOPRIGHT", module.mobFilter, "BOTTOMRIGHT")
module.mobFilter.deleteButton:SetText(L["Delete mob"])
module.mobFilter.deleteButton:Disable()
module.mobFilter.deleteButton:SetScript("OnClick", function(self)
	local scrollFrame = module.mobFilter.scrollFrame
	local btns = module.mobFilter.buttons
	if scrollFrame.selected then
		local selectedMob = CritlineMobFilter[scrollFrame.selected]
		tremove(CritlineMobFilter, scrollFrame.selected)
		local prevHighlight = btns[scrollFrame.selected - FauxScrollFrame_GetOffset(scrollFrame)]
		if prevHighlight then
			prevHighlight:UnlockHighlight()
		end
		scrollFrame.selected = nil
		scrollFrame:Update()
		self:Disable()
		core:Message(L["%s removed from mob filter."]:format(selectedMob))
	end
end)


do
	UIDropDownMenu_Initialize(module.filterType, function(self)
		local items = {
			{
				text = L["Spell filter"],
				value = 1,
				func = function()
					UIDropDownMenu_SetSelectedValue(self, 1)
					module.mobFilter:Hide()
					module.spellFilter:Show()
				end
			},
			{
				text = L["Mob filter"],
				value = 2,
				func = function()
					UIDropDownMenu_SetSelectedValue(self, 2)
					module.spellFilter:Hide()
					module.mobFilter:Show()
				end
			},
		}
		for _, item in ipairs(items) do
			UIDropDownMenu_AddButton(item)
		end
	end)

	UIDropDownMenu_SetWidth(module.filterType, 120)
	UIDropDownMenu_SetSelectedValue(module.filterType, 1)
end

do
	UIDropDownMenu_Initialize(module.spellFilter.tree, function(self)
		local items = {
			{
				text = DAMAGE,
				value = "dmg",
				func = function()
					local scrollFrame = module.spellFilter.scrollFrame
					UIDropDownMenu_SetSelectedValue(self, "dmg")
					FauxScrollFrame_SetOffset(scrollFrame, 0)
					_G[scrollFrame:GetName().."ScrollBar"]:SetValue(0)
					scrollFrame:Update()
				end
			},
			{
				text = HEALS,
				value = "heal",
				func = function()
					local scrollFrame = module.spellFilter.scrollFrame
					UIDropDownMenu_SetSelectedValue(self, "heal")
					FauxScrollFrame_SetOffset(scrollFrame, 0)
					_G[scrollFrame:GetName().."ScrollBar"]:SetValue(0)
					scrollFrame:Update()
				end
			},
			{
				text = PET,
				value = "pet",
				func = function()
					local scrollFrame = module.spellFilter.scrollFrame
					UIDropDownMenu_SetSelectedValue(self, "pet")
					FauxScrollFrame_SetOffset(scrollFrame, 0)
					_G[scrollFrame:GetName().."ScrollBar"]:SetValue(0)
					scrollFrame:Update()
				end
			},
		}
		for _, item in ipairs(items) do
			UIDropDownMenu_AddButton(item)
		end
	end)
	
	UIDropDownMenu_SetWidth(module.spellFilter.tree, 120)
	UIDropDownMenu_SetSelectedValue(module.spellFilter.tree, "dmg")
end


function SpellButton_OnModifiedClick(self, drag, ...)
	if module.spellFilter:IsVisible() and module:GetParent() and IsShiftKeyDown() then
		local spellID = SpellBook_GetSpellID(self:GetID())
		if spellID > MAX_SPELLS then
			return
		end
		module:AddSpell(UIDropDownMenu_GetSelectedValue(module.spellFilter.tree), (GetSpellName(spellID, SpellBookFrame.bookType)))
		return
    end
    OldSpellButton_OnModifiedClick(self, drag, ...)
end


function module.spellFilter.scrollFrame:Update()
	local selectedTree = UIDropDownMenu_GetSelectedValue(module.spellFilter.tree)
	
	local sortedSpells = {}
	for spell in pairs(CritlineSpellFilter[selectedTree]) do
		tinsert(sortedSpells, spell)
	end
	for spell in pairs(CritlineDB[selectedTree]) do
		if not CritlineSpellFilter[selectedTree][spell] then
			tinsert(sortedSpells, spell)
		end
	end
	sort(sortedSpells)
	
	local size = #sortedSpells
	
	FauxScrollFrame_Update(self, size, numSpellFilterButtons, 22)
	
	local offset = FauxScrollFrame_GetOffset(self)
	local button, lineplusoffset
	
	for line = 1, numSpellFilterButtons do
		button = module.spellFilter.buttons[line]
		lineplusoffset = line + offset
		if lineplusoffset <= size then
			button:SetText(sortedSpells[lineplusoffset])
			button:SetChecked(CritlineSpellFilter[selectedTree][sortedSpells[lineplusoffset]])
			button:Show()
		else
			button:Hide()
		end
	end
end


function module.mobFilter.scrollFrame:Update()
	local size = #CritlineMobFilter
	
	FauxScrollFrame_Update(self, size, numMobFilterButtons, 16)
	
	local offset = FauxScrollFrame_GetOffset(self)
	local button, lineplusoffset
	
	for line = 1, numMobFilterButtons do
		button = module.mobFilter.buttons[line]
		lineplusoffset = line + offset
		if lineplusoffset <= size then
			button:SetText(CritlineMobFilter[lineplusoffset])
			button:Show()
			if self.selected then
				if self.selected - FauxScrollFrame_GetOffset(self) == button:GetID() then
					button:LockHighlight()
				else
					button:UnlockHighlight()
				end
			end
		else
			button:Hide()
		end
	end
end


function module:AddSpell(tree, spell)
	if not CritlineSpellFilter[tree][spell] then
		CritlineSpellFilter[tree][spell] = true
		self.spellFilter.scrollFrame:Update()
		core:RebuildAllTooltips()
	end
end


function module:DeleteSpell(tree, spell)
	CritlineSpellFilter[tree][spell] = nil
	self.spellFilter.scrollFrame:Update()
	core:RebuildAllTooltips()
end


function module:IsMobInFilter(mobName, GUID)
	if GUID then
		if specialMobs[tonumber(strsub(GUID, 6, 12), 16)] then
			core:Debug("Mob ("..mobName..") is in default filter.")
			return true
		end
	end
	for i, v in ipairs(CritlineMobFilter) do
		if v == mobName then
			core:Debug("Mob ("..mobName..") is in custom filter.")
			return true
		end
	end
	-- core:Debug("Mob ("..mobName..") is not in filter.")
	return false
end


function module:HasSpecialAura()
	for i = 1, 40 do
		local buffName, _, buffIcon = UnitBuff("player", i)
		local debuffName, _, debuffIcon = UnitDebuff("player", i)
		for _, v in ipairs(Critline_SpecialAuras) do
			local name, _, icon = GetSpellInfo(v)
			if (buffName == name and buffIcon == icon) or (debuffName == name and debuffIcon == icon) then
				return true
			end
		end
	end
	return false
end


function module:SpellPassesFilter(tree, spell)
	local inFilter = CritlineSpellFilter[tree][spell]
	local isInverted = CritlineSettings[tree.."Invert"]
	return not (isInverted or inFilter) or (isInverted and inFilter)
end


function module:IsSpellInSpellbook(spellname)
	local numspells = 1
	while GetSpellName(numspells, BOOKTYPE_SPELL) do -- find the top rank of every spell
		local name = GetSpellName(numspells, BOOKTYPE_SPELL)
		if name == spellname then
			return true
		end
		numspells = numspells + 1
	end
	return false
end


function CritlineSpellFilterButton_OnClick(self)
	if self:GetChecked() then
		module:AddSpell(UIDropDownMenu_GetSelectedValue(module.spellFilter.tree), self:GetText())
	else
		module:DeleteSpell(UIDropDownMenu_GetSelectedValue(module.spellFilter.tree), self:GetText())
	end
end


function CritlineFiltersButton_OnClick(self)
	local scrollFrame = module.mobFilter.scrollFrame
	
	if scrollFrame.selected then
		if (scrollFrame.selected - FauxScrollFrame_GetOffset(scrollFrame)) == self:GetID() then
			-- clicking the selected button, clear selection
			self:UnlockHighlight()
			scrollFrame.selected = nil
		else
			-- clear selection if visible, and set new selection
			local prevHighlight = module.mobFilter.buttons[scrollFrame.selected - FauxScrollFrame_GetOffset(scrollFrame)]
			if prevHighlight then
				prevHighlight:UnlockHighlight()
			end
			self:LockHighlight()
			scrollFrame.selected = self:GetID() + FauxScrollFrame_GetOffset(scrollFrame)
		end
	else
		-- no previous selection, just set new and lock highlight
		self:LockHighlight()
		scrollFrame.selected = self:GetID() + FauxScrollFrame_GetOffset(scrollFrame)
	end
	
	-- enable/disable "Delete" button depending on if selection exists
	if scrollFrame.selected then
		module.mobFilter.deleteButton:Enable()
	else
		module.mobFilter.deleteButton:Disable()
	end
end